const std = @import("std");
const util = @import("util.zig");
const mem = std.mem;

pub fn main() !void {
    const input = @embedFile("data/day05.txt");
    const sln = try solve(util.gpa, input);
    std.debug.print("{d}\n", .{sln});
    std.debug.print("{d}\n", .{sln});
}

fn Range(comptime T: type) type {
    return struct {
        start: T,
        end: T,

        const Self = @This();

        /// Returns the intersection of two ranges.
        pub fn intersect(self: Self, other: Self) ?Self {
            const a = if (self.start < other.start) self else other;
            const b = if (self.start < other.start) other else self;

            if (a.start == a.end or b.start == b.end) return null; // Points don't intersect.

            if (a.end < b.start) return null; // No intersection.
            if (a.end > b.end) return b; // A is a superset of B.
            return .{ .start = b.start, .end = a.end };
        }

        /// Returns the union of two ranges.
        pub fn join(self: Self, other: Self) ?Self {
            if (self.intersect(other) == null) return null; // Can't union disjoint ranges.
            // Now guaranteed that they overlap.
            return .{ .start = @min(self.start, other.start), .end = @max(self.end, other.end) };
        }
    };
}

const Map = struct {
    offset: isize,
    source: Range(usize),
};

const Component = struct {
    stage: usize,
    range: Range(usize),
};

fn mergeRanges(ranges: *std.ArrayList(Component)) void {
    var i: usize = 0;
    while (i < ranges.items.len) : (i += 1) {
        var item = ranges.items[i];
        var j = i + 1;
        while (j < ranges.items.len) {
            const next = ranges.items[j];
            if (item.range.join(next.range)) |merged| {
                const merged_item = .{ .stage = item.stage, .range = merged };
                item = merged_item;
                ranges.items[i] = merged_item;
                _ = ranges.swapRemove(j);
            } else {
                j += 1;
            }
        }
    }
}

fn solve(alloc: mem.Allocator, input: []const u8) !usize {
    var it = util.splitLines(input);
    const seed_line = it.next().?;
    const seeds = try util.parseIntsScalar(usize, alloc, seed_line[mem.indexOfScalar(u8, seed_line, ':').? + 1 ..], .{});
    defer alloc.free(seeds);

    var ranges = std.ArrayList(Component).init(alloc);
    defer ranges.deinit();

    {
        var index: usize = 0;
        while (index < seeds.len) : (index += 2) {
            try ranges.append(.{
                .stage = 0,
                .range = .{
                    .start = seeds[index],
                    .end = seeds[index] + seeds[index + 1],
                },
            });
        }
    }

    var current_stage: usize = 0;
    while (it.next()) |line| {
        if (line.len == 0) {
            for (ranges.items, 0..) |item, index| {
                if (item.stage != current_stage) {
                    ranges.items[index] = .{ .stage = current_stage, .range = item.range };
                }
            }
            mergeRanges(&ranges);
        } else if (mem.endsWith(u8, line, "map:")) {
            current_stage += 1;
        } else {
            const map_line = try util.parseIntsScalar(usize, alloc, line, .{});
            defer alloc.free(map_line);
            const map = Map{
                .offset = @as(isize, @intCast(map_line[0])) - @as(isize, @intCast(map_line[1])),
                .source = .{ .start = map_line[1], .end = map_line[1] + map_line[2] },
            };

            var i: usize = ranges.items.len;
            while (i != 0) : (i -= 1) {
                if (ranges.items[i - 1].stage == current_stage) continue;
                const item = ranges.swapRemove(i - 1);

                if (item.range.intersect(.{ .start = 0, .end = map.source.start })) |section| {
                    // Non-transitioning section.
                    try ranges.append(.{ .stage = item.stage, .range = section });
                }
                if (item.range.intersect(map.source)) |section| {
                    // Transitioning section.
                    try ranges.append(.{
                        .stage = current_stage,
                        .range = .{
                            .start = @intCast(@as(isize, @intCast(section.start)) + map.offset),
                            .end = @intCast(@as(isize, @intCast(section.end)) + map.offset),
                        },
                    });
                }
                if (item.range.intersect(.{ .start = map.source.end, .end = std.math.maxInt(usize) })) |section| {
                    // Non-transitioning section.
                    try ranges.append(.{ .stage = item.stage, .range = section });
                }
            }
        }
    }

    var min_loc: usize = std.math.maxInt(usize);
    for (ranges.items) |item| {
        min_loc = @min(min_loc, item.range.start);
    }
    return min_loc;
}

test "silver" {
    const input =
        \\seeds: 79 14 55 13
        \\
        \\seed-to-soil map:
        \\50 98 2
        \\52 50 48
        \\
        \\soil-to-fertilizer map:
        \\0 15 37
        \\37 52 2
        \\39 0 15
        \\
        \\fertilizer-to-water map:
        \\49 53 8
        \\0 11 42
        \\42 0 7
        \\57 7 4
        \\
        \\water-to-light map:
        \\88 18 7
        \\18 25 70
        \\
        \\light-to-temperature map:
        \\45 77 23
        \\81 45 19
        \\68 64 13
        \\
        \\temperature-to-humidity map:
        \\0 69 1
        \\1 0 69
        \\
        \\humidity-to-location map:
        \\60 56 37
        \\56 93 4
    ;
    _ = input;
    // const sln = try solve(std.testing.allocator, input);
    // try std.testing.expectEqual(@as(usize, 35), sln);
}

test "gold" {
    const input =
        \\seeds: 79 14 55 13
        \\
        \\seed-to-soil map:
        \\50 98 2
        \\52 50 48
        \\
        \\soil-to-fertilizer map:
        \\0 15 37
        \\37 52 2
        \\39 0 15
        \\
        \\fertilizer-to-water map:
        \\49 53 8
        \\0 11 42
        \\42 0 7
        \\57 7 4
        \\
        \\water-to-light map:
        \\88 18 7
        \\18 25 70
        \\
        \\light-to-temperature map:
        \\45 77 23
        \\81 45 19
        \\68 64 13
        \\
        \\temperature-to-humidity map:
        \\0 69 1
        \\1 0 69
        \\
        \\humidity-to-location map:
        \\60 56 37
        \\56 93 4
    ;
    const sln = try solve(std.testing.allocator, input);
    try std.testing.expectEqual(@as(usize, 46), sln);
}
